#!/usr/bin/env python
'''
Software Heritage functions.
'''

import logging
from urllib.parse import quote as urlquote

import requests


LOGGER = logging.getLogger(__name__)


def get_swhid_by_origin(origin):
    '''
    Get a Software Heritage ID from an origin URL. This only includes the origin
    and repository ID.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        A Software Heritage ID string.
    '''
    cache = get_swhid_by_origin.cache
    try:
        return cache[origin]
    except KeyError:
        pass

    timeout = 5
    url = f'https://archive.softwareheritage.org/api/1/origin/{origin}/get/'
    LOGGER.debug('Querying %s', url)
    resp = requests.get(url, timeout=timeout).json()
    try:
        resp = requests.get(resp['origin_visits_url'], timeout=timeout).json()
        resp = requests.get(resp[0]['snapshot_url'], timeout=timeout).json()
        resp = requests.get(resp['branches']['HEAD']['target_url'], timeout=timeout).json()
        swhid = f'swh:1:dir:{resp["directory"]};origin={urlquote(origin)}'
        cache[origin] = swhid
        return swhid
    except KeyError:
        return None


get_swhid_by_origin.cache = {}


def get_swh_url_by_origin(origin):
    '''
    Get a Software Heritage link for an origin URL.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        A Software Heritage URL. The page may not exist.
    '''
    return f'https://archive.softwareheritage.org/browse/origin/?origin_url={urlquote(origin)}'


def swh_project_exists(origin):
    '''
    Check if the project exists on Software Heritage.

    Args:
        origin:
            The origin URL of the software project.

    Returns:
        True if the project exists, False otherwise.
    '''
    return get_swhid_by_origin(origin) is not None



#!/usr/bin/env python
'''\
Update project metadata.\
'''

import argparse
import logging
import pathlib
import subprocess
import sys

from .common import NAME
from .config import CONFIG_FILE, CONFIG_EXAMPLE
from .project import Project


LOGGER = logging.getLogger(__name__)


def main(args=None):
    '''
    Main function.
    '''
    parser = argparse.ArgumentParser(
        description=__doc__,
        prog=NAME
    )
    parser.add_argument(
        'path',
        nargs='*',
        help='Path to project directory.'
    )
    parser.add_argument(
        '--gen-config',
        metavar='PATH',
        help=f'''
            Generate a configuration file template at the given path. If the
            path is "-", the file will be printed to STDOUT. Note that %(prog)s
            will only look for files named {CONFIG_FILE}.
            '''
    )
    parser.add_argument(
        '--trust',
        action='store_true',
        help='''
            It is possible to insert arbitrary command output into the README
            file. By default, %(prog)s will prompt the user for confirmation
            before running the command to prevent arbitrary code execution in
            the context of a collaborative environment. This option can be used
            to disable the prompt if the user trusts all of the commands in the
            README.
            '''
    )
    pargs = parser.parse_args(args=args)

    if pargs.gen_config:
        if pargs.gen_config == '-':
            print(CONFIG_EXAMPLE)
        else:
            path = pathlib.Path(pargs.gen_config).resolve()
            LOGGER.info('Creating %s', path)
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(CONFIG_EXAMPLE, encoding='utf-8')

    for path in pargs.path:
        LOGGER.info('Updating %s', path)
        proj = Project(path, trust_commands=pargs.trust)
        proj.update()


def run_main(args=None):
    '''
    Wrapper around main for exception handling.
    '''
    logging.basicConfig(
        style='{',
        format='[{asctime}] {levelname} {message}',
        level=logging.INFO
    )
    try:
        main(args=args)
    except KeyboardInterrupt:
        pass
    except subprocess.CalledProcessError as err:
        sys.exit(err)


if __name__ == '__main__':
    run_main()



#!/usr/bin/env python3
'''
Base class for inserting content into text files.
'''

import pathlib
import re


from .file import update_content


class Inserter():
    '''
    Insert updated content into a text file.
    '''
    def __init__(self, regex):
        '''
        Args:
            regex:
                A compiled regular expression.
        '''
        self.regex = regex

    def insert(self, match):
        '''
        Given a regular expression match, return the content to insert.

        Args:
            match:
                The Match object.

        Returns:
            The string to substitute for the match.
        '''
        return match.group(0)

    def update(self, path, encoding='utf-8'):
        '''
        Insert the configured content into the file.
        '''
        text = pathlib.Path(path).resolve().read_text(encoding=encoding)
        text = self.regex.sub(self.insert, text)
        update_content(text, path)


class MarkdownInserter(Inserter):
    '''
    Inserter with custom regex for invisible comments in Markdown files.

    In Markdown, a link label for a simple hash results in no output for most
    common Markdown processors: "[comment]: #". Note that these must be
    proceeded by blank lines to be recognized in Markdown.

    This class recognized blocks that begin with a comment prefixed with
    "insert: " and end with a comment containing only "/insert". Both comments
    must have the same indentation level and the inserted text will be indented
    to the same level.
    '''
    def __init__(self):
        regex = re.compile(
            r'^(?P<indent>[ \t]*)\[insert: (?P<label>.+?)\]: #\n'
            r'(?:(?P<content>.*?)\n)?'
            r'^(?P=indent)\[/insert: (?P=label)]: #\n',
            re.MULTILINE | re.DOTALL
        )
        super().__init__(regex)

    @staticmethod
    def get_link(label, url):
        '''
        Get the Markdown link.

        Args:
            label:
                The label shown for the link.

            url:
                The URL of the link.

        Return:
            The Markdown link as a string.
        '''
        return f'[{label}]({url})'

    def get_output(self, label, content):  # pylint: disable=unused-argument
        '''
        Get output to replace the given label. Override this to insert custom
        content.

        Args:
            label:
                The label in the comment pair that was matched.

            content:
                The content between the matched comments.

        Returns:
            The content to insert, or None if the label was not recognized.
        '''
        return None

    def insert(self, match):
        indent = match['indent']
        label = match['label'].strip()
        content = match['content']
        output = self.get_output(label, content)  # pylint: disable=assignment-from-none
        if output is None:
            return match.group(0)
        output = '\n'.join(f'{indent}{line}' for line in output.strip('\n').split('\n'))
        return f'{indent}[insert: {label}]: #\n\n{output}\n\n{indent}[/insert: {label}]: #\n'



#!/usr/bin/env python
'''
Virtual environment functions.
'''

import contextlib
import logging
import pathlib
import sysconfig
import tempfile
import subprocess
import sys


LOGGER = logging.getLogger(__name__)


@contextlib.contextmanager
def virtual_environment(update_pip=False, inherit=False):
    '''
    Context manager for creating a temporary virtual environment.

    Args:
        update_pip:
            If True, update pip in the virtual environment.

        inherit:
            If True, create a .pth file to inherit packages from the parent
            environment.
    '''
    parent_purelib = sysconfig.get_paths()['purelib']

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        venv_dir = tmp_dir / 'venv'
        python_exe = pathlib.Path(sys.executable).resolve()
        cmd = [str(python_exe), '-m', 'venv', str(venv_dir)]
        LOGGER.debug('Creating temporary virtual environment: %s', cmd)
        subprocess.run(cmd, check=True)
        venv_exe = venv_dir / 'bin' / python_exe.name
        if update_pip:
            cmd = [str(venv_exe), '-m', 'pip', 'install', '-U', 'pip']
            LOGGER.debug('Updating pip in virtual environment: %s', cmd)
            subprocess.run(cmd, check=True)
        if inherit:
            cmd = [
                str(venv_exe), '-c',
                'import sysconfig; print(sysconfig.get_paths()["purelib"])'
            ]
            LOGGER.debug("Locating virtual environment's purelib directory.")
            child_purelib = subprocess.run(
                cmd, check=True, stdout=subprocess.PIPE
            ).stdout.decode().strip()
            pth_path = pathlib.Path(child_purelib) / 'parent.pth'
            pth_path.write_text(parent_purelib, encoding='utf-8')
        yield venv_dir, venv_exe



#!/usr/bin/env python
'''
All things specific to Python (pyproject.toml, PyPI, etc.)
'''

import logging
import pathlib
import subprocess
import tomllib
from urllib.parse import quote as urlquote

import requests
import tomli_w
import tomli_w._writer
from spdx_license_list import LICENSES as SPDX_LICENSES
from trove_classifiers import classifiers as py_classifiers

from ..file import update_content
from .venv import virtual_environment

LOGGER = logging.getLogger(__name__)


def get_pypi_url(name):
    '''
    Get the URL to the project's page on PyPI if it exists.

    Args:
        name:
            The project name.

    Returns:
        The project URL, or None if it does not exist.
    '''
    timeout = 5
    url = f'https://pypi.org/pypi/{urlquote(name)}/json'
    LOGGER.debug('Querying %s', url)
    resp = requests.get(url, timeout=timeout).json()
    try:
        return resp['info']['package_url']
    except KeyError:
        return None


def get_license_classifier(spdx_id):
    '''
    Get the Python license classifier for the given SPDX license ID.

    Args:
        spdx_id:
            The SPXD license ID.

    Returns:
        The corresponding Python trove classifier if found, else None.
    '''
    spdx_data = SPDX_LICENSES.get(spdx_id)
    if not spdx_data:
        LOGGER.warning('%s is not a recognized SPDX license ID.', spdx_id)
        return None
    name = spdx_data.name
    if spdx_data.osi_approved:
        expected = f'License :: OSI Approved :: {name}'
    else:
        expected = 'License :: {name}'
    if expected in py_classifiers:
        return expected
    closest = None
    for classifier in py_classifiers:
        if classifier.startswith(expected):
            if closest is None or len(closest) > len(classifier):
                closest = classifier
    if closest is None:
        LOGGER.warning(
            'Failed to map SPDX license ID %s to a Python trove classifier.',
            spdx_id
        )
    return closest


def update_pyproject_toml(project):
    '''
    Update the URLs in a pyproject.toml file.

    Args:
        project:
            The Project instance.
    '''
    path = project.pyproject_toml_path
    data = tomllib.loads(path.read_text(encoding='utf-8'))

    urls = data['project']['urls']
    urls.clear()
    urls.update(project.urls)

    data['project']['authors'] = [
        {
            'name': f'{author["given-names"]} {author["family-names"]}',
            'email': author['email']
        }
        for author in project.config.config['authors']
    ]

    classifiers = set(
        classifier
        for classifier in data['project'].get('classifiers', [])
        if not classifier.startswith('License :: ')
    )

    spdx_id = project.spdx_license
    if spdx_id:
        license_classifier = get_license_classifier(spdx_id)
        if license_classifier:
            classifiers.add(license_classifier)
    else:
        LOGGER.warning('No license detected for project.')

    data['project']['classifiers'] = sorted(classifiers)

    content = tomli_w.dumps(data, multiline_strings=True)
    update_content(content, path)
    return urls


def get_version(project_dir):
    '''
    Get the version of a project from its directory.

    This will create a temporary virtual environment and install the project in
    it to get the version. This ensures that VCS-versions are correctly handled.

    This should not be necessary but at the time or writing the current version
    of CodeMetaPy fails to detect versions.
    '''
    project_dir = pathlib.Path(project_dir).resolve()
    with virtual_environment() as (_venv_dir, python_exe):
        cmd = [python_exe, '-m', 'pip', 'install', '--no-deps', '-U', str(project_dir)]
        LOGGER.debug('Installing package in virtual environment: %s', cmd)
        subprocess.run(cmd, check=True)

        data = tomllib.loads((project_dir / 'pyproject.toml').read_text(encoding='utf-8'))
        name = data['project']['name']
        cmd = [
            python_exe, '-c',
            f'from importlib.metadata import version; print(version("{name}"))'
        ]
        LOGGER.debug('Getting version of %s: %s', name, cmd)
        return subprocess.run(cmd, check=True, capture_output=True).stdout.decode().strip()



#!/usr/bin/env python
'''
Citation file format (.cff) functions.

https://citation-file-format.github.io/
https://github.com/citation-file-format/citation-file-format/blob/main/schema-guide.md
'''


import json

import yaml

from .file import update_content
from .id.hal import get_hal_url_by_origin, get_hal_id_from_url
from .id.swh import get_swhid_by_origin, get_swh_url_by_origin


TEMPLATE = '''\
cff-version: 1.2.0
message: If you use this software, please cite it using these metadata.
title: My Research Software
abstract: This is my awesome research software. It does many things.
authors:
  - family-names: Rye
    given-names: Jan-Michael
    email: jan-michael.rye@inria.fr
    orcid: "https://orcid.org/0009-0005-0109-6598"
version: 0.11.2
# date-released: "yyyy-mm-dd"
identifiers:
  - description: Collection of archived snapshots of My Research Software
    type: doi
    value: "10.5281/zenodo.123456"
  - description: Archived snapshot of version 0.11.2 of My Research Software
    type: doi
    value: "10.5281/zenodo.123457"
license: MIT
repository-code: "https://github.com/citation-file-format/my-research-software"
'''


class Citation():
    '''
    Citation data manager.
    '''
    def __init__(self, project):
        '''
        Args:
            project:
                A Project instance.
        '''
        self.project = project

    @staticmethod
    def get_template():
        '''
        Get the cff file template as an object.
        '''
        return yaml.safe_load(TEMPLATE)

    @staticmethod
    def _transform_codemeta_authors(authors):
        '''
        Transform the CodeMeta authors list to citation file format.
        '''
        for author in authors:
            data = {}
            for cff_key, cm_key in (
                ('family-names', 'familyName'),
                ('given-names', 'givenName'),
                ('email', 'email')
            ):
                data[cff_key] = author[cm_key]
            author_id = author.get('@id')
            if author_id and '//orcid.org/' in author_id:
                data['orcid'] = author_id
            yield data

    def update(self):
        '''
        Update the citation file with data from codemeta.json and other sources.

        Returns:
            The citation file object.
        '''
        path = self.project.citation_cff_path
        if not path.exists() and not self.project.config.get('citations_cff', default=False):
            return
        cff_data = self.get_template()
        with self.project.codemeta_json_path.open('rb') as handle:
            cm_data = json.load(handle)
        for cff_key, cm_key in (
            ('title', 'name'),
            ('abstract', 'description'),
            ('version', 'version'),
            ('repository-code', 'codeRepository')
        ):
            try:
                cff_data[cff_key] = cm_data[cm_key]
            except KeyError:
                del cff_data[cff_key]

        try:
            cff_data['license'] = cm_data['license'].rsplit('/', 1)[1]
        except KeyError:
            del cff_data['license']
        cff_data['authors'] = list(self._transform_codemeta_authors(cm_data['author']))

        identifiers = cff_data['identifiers']
        identifiers.clear()
        git_url = self.project.git_repo.public_git_url
        swh_id = get_swhid_by_origin(git_url)
        if swh_id:
            # TODO
            # Uncomment once cff convert supports this.
            #  identifiers.append({
            #      'type': 'swh',
            #      'value': swh_id,
            #      'description': 'The Software Heritage identifier.'
            #  })
            identifiers.append({
                'type': 'url',
                'value': get_swh_url_by_origin(git_url),
                'description': 'The Software Heritage URL.'
            })
            hal_url = get_hal_url_by_origin(git_url)
            if hal_url:
                identifiers.append({
                    'type': 'url',
                    'value': hal_url,
                    'description': 'HAL open science URL.'
                })
                identifiers.append({
                    'type': 'other',
                    'value': get_hal_id_from_url(hal_url),
                    'description': 'HAL open science identifier.'
                })

        content = yaml.dump(cff_data)
        update_content(content, path)



#!/usr/bin/env python
'''
Configuration.
'''


import logging
import os
import pathlib
import shlex

import gitlab
import yaml

from .common import NAME

LOGGER = logging.getLogger(__name__)
CONFIG_FILE = f'{NAME}.yaml'

CONFIG_EXAMPLE = '''
# A list of authors. They will appear in various files (e.g. pyproject.toml,
# codemeta.json, CITATIONS.cff).
authors:
    # Given names (required)
  - given-names: John

    # Family names (required)
    family-names: Doe

    # Email (optional)
    email: john.doe@example.com

    # Affiliation (optional)
    affiliation: Example Institute

    # ORCID identifier (optional)
    orcid: XXXX-XXXX-XXXX-XXXX

    # HAL Open Science identifier (optional)
    hal: XXXXXXX

# If true, create missing CITATIONS.cff files.
citations_cff: true

# GitLab settings (optional)
gitlab:
  # Prometa uses python-gitlab to manage GitLab hooks that push code to other
  # open repositories (currently only Software Heritage). python-gitlab requires
  # both a configuration file and the name of the section in the configuration
  # file to use for a given project. For details, see the documentation:
  #
  # https://python-gitlab.readthedocs.io/en/stable/cli-usage.html#cli-configuration
  #
  # python-gitlab configuration file:
  config: path/to/python-gitlab.cfg

  # The section of the python-gitlab configuration file to use when retrieving
  # GitLab project data.
  section: somewhere

  # Map GitLab hosts to their corresponding GitLab Pages URL formats. This map
  # will be used to generate documentation links when a "pages" job is detected
  # in the CI configuration file. The namespace and name parameters correspond
  # to those of the GitLab project.
  pages_urls:
    gitlab.com: "https://{namespace}.gitlab.io/{name}"

  # The regular expression for matching release tags. If given, a CI release job
  # will be created for tags that match this pattern. Omit this or set it to
  # null to disable release jobs.
  release_tag_regex: "^v."

  # Configure tags for GitLab CI jobs. This is a mapping of Python regular
  # expressions to lists of tags. Jobs that match the regular expressions will
  # be tagged with the corresponding tags. If multiple regular expressions match
  # a job then it will accumulate the tags.
  #
  # To apply the same tags to all jobs, use the regular expression ".".
  ci_tags:
      ".":
        - tag1
        - tag2
        - tag3

# The utility to use when merging changes. It must accept two file paths (the
# modified file and the original) and return non-zero exit status to indicate an
# error or abort.
merger: vimdiff
'''.strip()


def _nested_values(data):
    '''
    A generator over all nested values in a dict.
    '''
    if not isinstance(data, dict):
        yield data
        return
    for value in data.values():
        if isinstance(value, dict):
            yield from _nested_values(value)
        else:
            yield value


def _nested_update(old_data, new_data, origin, new_path):
    '''
    Updated nested values in a configuration dict recursively.

    Args:
        old_data:
            The dict to update in place.

        new_data:
            The data to add to old_dict.

        origin:
            A dict of the same layout as old_data but the values are updated
            with the filenames that provided them.

        new_path:
            The path from which new_data was loaded.
    '''
    for key, new_value in new_data.items():
        old_value = old_data.get(key)
        if isinstance(new_value, dict) and isinstance(old_value, dict):
            _nested_update(old_value, new_value, origin[key], new_path)
        else:
            old_data[key] = new_value
            origin[key] = new_path


class Config():
    '''
    Common non-derivable configuration.
    '''
    def __init__(self, proj_path):
        '''
        Args:
            proj_path:
                The project path.
        '''
        self.proj_path = pathlib.Path(proj_path).resolve()
        self._config = None
        self._origin = None

    @property
    def config_paths(self):
        '''
        A generator over configuration file paths. The files are located by
        scanning the project directory and then all parent directories, with
        files closer to the project directory taking precedence. This allows
        common settings to be placed in a common root directory.
        '''
        path = self.proj_path / CONFIG_FILE
        while True:
            if path.exists():
                LOGGER.info('Found configuration file: %s', path)
                yield path
            next_path = path.parent.parent / CONFIG_FILE
            if path != next_path:
                path = next_path
            else:
                return

    @property
    def config(self):
        '''
        The configuration file object. If None, there is no configuration file.
        '''
        if self._config is None:
            configs = [
                (path, yaml.safe_load(path.read_text(encoding='utf-8')))
                for path in self.config_paths
            ]
            final_config = {}
            origin = {}
            for path, config in reversed(configs):
                _nested_update(final_config, config, origin, path)
            self._config = final_config
            self._origin = origin
        return self._config

    def get(self, *keys, default=None):
        '''
        Retrieve a configuration file value. This will scan the loaded
        configuration files in order and return the first match.

        Args:
            *keys:
                The keys to the field. For example, to retrieve the value of
                "bar" under "foo", call get("foo", "bar"). Integers may also be
                used to index lists.

            default:
                The default value to return if no value was found.

        Returns:
            The target value, or the default if no value was found.
        '''
        config = self.config
        origin = self._origin
        for i, key in enumerate(keys):
            try:
                config = config[key]
                try:
                    origin = origin[key]
                except TypeError:
                    pass
            except (KeyError, IndexError):
                return default
            except TypeError as err:
                paths = sorted(set(_nested_values(origin)))
                LOGGER.error(
                    'Failed to retrieve configuration value for %s: %s [%s]',
                    keys[:i + 1],
                    err,
                    ', '.join(shlex.quote(str(p)) for p in paths)
                )
                return default
        return config

    def _get_origin(self, *keys):
        '''
        Similar to get() but returns the origin path for the given keys.
        '''
        origin = self._origin
        if isinstance(origin, pathlib.Path):
            return origin
        for key in keys:
            origin = origin[key]
            if isinstance(origin, pathlib.Path):
                return origin
        return None

    @property
    def gitlab(self):
        '''
        The python-gitlab GitLab instance from the current configuration.
        '''
        path = self.get('gitlab', 'config')
        if path is None:
            LOGGER.warning('No python-gitlab configuration file specified.')
            path = os.getenv('PYTHON_GITLAB_CFG')
            if path is None:
                LOGGER.warning('PYTHON_GITLAB_CFG environment variable is unset.')
                return None
            path = pathlib.Path(path)
        else:
            path = self._get_origin('gitlab', 'config').parent.joinpath(path)
        path = path.resolve()
        section = self.get('gitlab', 'section')
        if not section:
            LOGGER.warning(
                'No section specified for the python-gitlab configuration file: %s',
                path
            )
        return gitlab.Gitlab.from_config(section, [str(path)])



#!/usr/bin/env python
'''
HAL open science functions.

https://api.archives-ouvertes.fr/docs/search
'''

import logging
from urllib.parse import quote as urlquote

import requests


LOGGER = logging.getLogger(__name__)


def get_hal_url_by_origin(origin):
    '''
    Get the HAL ID by origin.
    '''
    cache = get_hal_url_by_origin.cache
    try:
        return cache[origin]
    except KeyError:
        pass

    timeout = 5
    query = origin.split('://', 1)[1]
    search_url = f'https://api.archives-ouvertes.fr/search/?q={urlquote(query)}'
    LOGGER.debug('Querying %s', search_url)
    resp = requests.get(search_url, timeout=timeout).json()
    for doc in resp['response']['docs']:
        label = doc['label_s']
        if f'origin={origin};' in label:
            url = doc['uri_s']
            cache[origin] = url
            return url
    return None


get_hal_url_by_origin.cache = {}


def get_hal_id_from_url(url):
    '''
    Extract the HAL ID from a HAL URL.
    '''
    return url.rsplit('/', 1)[1]



#!/usr/bin/env python3
'''
Insert data into the README.
'''

import logging
import shlex
import subprocess
import tomllib

from .id.hal import get_hal_url_by_origin
from .id.swh import get_swh_url_by_origin, swh_project_exists
from .insert import MarkdownInserter
from .python.common import get_pypi_url


LOGGER = logging.getLogger(__name__)


class ReadmeInserter(MarkdownInserter):
    '''
    Insert data into the Markdown file.
    '''
    def __init__(self, project):
        super().__init__()
        self.project = project
        self._labels_found = set()

    @staticmethod
    def _split_level(label):
        '''
        Extract the header level from a label.

        Args:
            label:
                A string with the label and header leavel in format "<label>
                <level>".

        Returns:
            The label and the level. If no level was found, the level defaults
            to 1.
        '''
        label = label.strip()
        try:
            label, level = label.split(None, 1)
        except ValueError:
            LOGGER.debug('Missing header level in label [%s]: assuming 1', label)
            return label, 1
        return label, int(level)

    @staticmethod
    def _get_lang_and_command(label):
        '''
        Extract command and optional language from the label.

        Args:
            label:
                A string of the format "command[:<lang>] <cmd>", e.g. "command
                echo test" or "command:yaml some_cmd_to_print_yaml". The command
                string will be parsed by shlex.split().

        Returns:
            The language and the command. The language will be None if not found
            and the command will be a list of command words.
        '''
        label, command = label.split(None, 1)
        try:
            label, lang = label.split(':', 1)
            lang = lang.strip()
        except ValueError:
            lang = None
        return lang, shlex.split(command)

    def _get_links_section(self, label):
        '''
        Get the section containing standard links.
        '''
        label, level = self._split_level(label)
        header_prefix = '#' * (level + 1)
        is_pyproject = self.project.pyproject_toml_path.exists()

        lines = [f'{header_prefix} GitLab', '']

        lines.extend(
            self.get_link(name.title(), link)
            for name, link in self.project.urls.items()
        )

        other_repos = []
        if is_pyproject:
            lines.append(self.get_link(
                'GitLab package registry',
                self.project.git_repo.get_section_url('packages')
            ))
            pyproj_data = tomllib.loads(
                self.project.pyproject_toml_path.read_text(encoding='utf-8')
            )
            pypi_url = get_pypi_url(pyproj_data['project']['name'])
            if pypi_url:
                other_repos.append(self.get_link('Python Package Index', pypi_url))

        origin_url = self.project.git_repo.public_git_url
        if swh_project_exists(origin_url):
            other_repos.append(
                self.get_link('Software Heritage', get_swh_url_by_origin(origin_url))
            )

        hal_url = get_hal_url_by_origin(origin_url)
        if hal_url:
            other_repos.append(self.get_link('HAL open science', hal_url))

        if other_repos:
            lines.extend((
                '',
                f'{header_prefix} Other Repositories',
                '',
                *other_repos
            ))

        return '\n'.join(
            line if not line or line.startswith('#') else f'* {line}'
            for line in lines
        )

    def _get_citation_section(self, _label):
        '''
        Get the section containing citation examples for different targets.
        '''
        path = self.project.citation_cff_path
        citation_cff_url = self.project.git_repo.get_main_blob_url(path.name)
        blocks = [
            'Please cite this software using the metadata provided in '
            f'[{path.name}]({citation_cff_url}). '
            'The following extracts are provided for different applications.',
            ''
        ]
        if path.exists():
            for fmt in ('cff', 'apalike', 'bibtex', 'endnote', 'ris', 'schema.org', 'zenodo'):
                cmd = ['cffconvert', '-i', str(path), '-f', fmt]
                LOGGER.debug('Converting %s to %s', path, fmt)
                output = subprocess.run(
                    cmd,
                    check=True,
                    stdout=subprocess.PIPE,
                ).stdout.decode()
                blocks.append(f'{fmt}\n: ~~~\n{output.strip()}\n~~~\n')
        return '\n'.join(blocks)

    def _get_command_output_section(self, label):
        '''
        Get the output of a command.
        '''
        lang, command = self._get_lang_and_command(label)
        if not self.project.trust_commands:
            ans = input(f'Run command {command}? [y/N] ')
            if ans.strip().lower() != 'y':
                LOGGER.info('Skipping command %s', command)
                return None
        LOGGER.info('Running command: %s', command)
        output = subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE
        ).stdout.decode()
        lang = lang if lang else ''
        return f'~~~{lang}\n$ {shlex.join(command)}\n{output}\n~~~\n'

    def get_output(self, label, content):
        '''
        Override parent get_output.
        '''
        self._labels_found.add(label)
        if label.startswith('citations'):
            return self._get_citation_section(label)
        if label.startswith('links'):
            return self._get_links_section(label)
        if label.startswith('command_output'):
            output = self._get_command_output_section(label)
            return content if output is None else output
        LOGGER.warning('Unhandled label in README: %s', label)
        return content



#!/usr/bin/env python
'''
Git functions.
'''

import logging
import pathlib
import subprocess


LOGGER = logging.getLogger(__name__)


class GitlabRepo():
    '''
    Basic functionality for retrieving Git information.
    '''

    def __init__(self, path, remote='origin'):
        '''
        Args:
            path:
                A path to the Git repository or any non-submodule path within
                it.

            remote:
                The remote repository name to use for generating URLs.
        '''
        self.path = pathlib.Path(path).resolve()
        self.path = self.top_level
        self.remote = remote
        self._origin_url = None

    def run_cmd(self, cmd):
        '''
        Run a git command and return its output.

        Args:
            cmd:
                The git sub-command and arguments.

        Returns:
            The command output.
        '''
        cmd = ('git', '-C', str(self.path), *cmd)
        LOGGER.debug('Running command: %s', cmd)
        return subprocess.run(cmd, check=True, capture_output=True).stdout.decode().strip()

    @property
    def top_level(self):
        '''
        The top-level directory.
        '''
        path = self.run_cmd(('rev-parse', '--show-toplevel'))
        return pathlib.Path(path).resolve()

    @property
    def remote_url(self):
        '''
        The origin URL.
        '''
        if self._origin_url is None:
            self._origin_url = self.run_cmd(('config', '--get', 'remote.origin.url'))
        return self._origin_url

    @property
    def parsed_origin(self):
        '''
        A 3-tuple of the origin's host, namespace and project name.
        '''
        host, subpath = self.remote_url.split(':', 1)
        subpath = subpath.rsplit('.', 1)[0]
        namespace, name = subpath.split('/', 1)
        host = host.split('@', 1)[1]
        return host, namespace, name

    @property
    def public_url(self):
        '''
        The URL for the main project page.
        '''
        host, namespace, name = self.parsed_origin
        return f'https://{host}/{namespace}/{name}'

    def get_section_url(self, section):
        '''
        The URL to one of the various GitLab sections, e.g. "blob/main" or
        "packages".
        '''
        section = section.lstrip()
        return f'{self.public_url}/-/{section}'

    @property
    def public_git_url(self):
        '''
        The publically accessible HTTPS URL.
        '''
        return f'{self.public_url}.git'

    def get_main_blob_url(self, path):
        '''
        Get the main branch URL to the given path.
        '''
        path = path.lstrip('/')
        return self.get_section_url(f'blob/main/{path}')

    @property
    def readme_url(self):
        '''
        The URL to the README.
        '''
        return self.get_main_blob_url('README.md')

    def get_gitlab_project(self, glab):
        '''
        Get the python-gitlab Project instance from a python-gitlab GitLab
        instance.
        '''
        _host, namespace, name = self.parsed_origin
        return glab.projects.get(f'{namespace}/{name}')



#!/usr/bin/env python
'''
CodeMeta functions.
'''

import contextlib
import json
import logging
import pathlib
import shutil
import subprocess
import tempfile
import tomllib

import tomli_w
import tomli_w._writer

from .file import diff
from .id.orcid import get_orcid_url
from .python.common import get_version
from .python.venv import virtual_environment


LOGGER = logging.getLogger(__name__)


# https://codemeta.github.io/user-guide/
class CodeMeta():
    '''
    Wrapper around the codemetapy command-line executable.
    '''

    def __init__(self, project):
        '''
        Args:
            project:
                The project object.
        '''
        self.project = project
        self._data = None

    @property
    def data(self):
        '''
        The CodeMeta data.
        '''
        if self._data is None:
            try:
                with self.project.codemeta_json_path.open('rb') as handle:
                    self._data = json.load(handle)
            except FileNotFoundError:
                return None
        return self._data

    @property
    def name(self):
        '''
        The project name.
        '''
        try:
            return self.data['name']
        except TypeError:
            return None

    @contextlib.contextmanager
    def _modified_pyproject_toml(self):
        '''
        Context manager to create a temporary modified pyproject.toml file that
        works around current bugs in CodeMetaPy. The origin of the problem is
        that the project metadata now returns custom objects for the "readme"
        and "license" fields while CodeMetaPy still expects these to be strings.
        '''
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            git_dir = self.project.git_repo.path
            shutil.copytree(git_dir, tmp_dir, dirs_exist_ok=True, symlinks=True)
            ppt_path = git_dir / 'pyproject.toml'
            tmp_ppt_path = tmp_dir / ppt_path.name

            ppt_data = tomllib.loads(ppt_path.read_text(encoding='utf-8'))
            for field in ('license', 'readme'):
                try:
                    del ppt_data['project'][field]
                except KeyError:
                    pass
            content = tomli_w.dumps(ppt_data, multiline_strings=True)
            tmp_ppt_path.write_text(content, encoding='utf-8')

            yield tmp_ppt_path, ppt_data['project']['name']

    def from_pyproject_toml(self, version=None):
        '''
        Update from a pyproject.toml file.

        Args:
            version:
                The version to set. This is necessary due to SCM incrementing
                the version for unclean directories.
        '''
        author_data = []
        for author in self.project.config.config['authors']:
            data = {
                '@type': 'Person',
                'email': author['email'],
                'givenName': author['given-names'],
                'familyName': author['family-names']
            }
            orcid = author.get('orcid')
            if orcid:
                data['@id'] = get_orcid_url(orcid)
            author_data.append(data)

        codemeta_json_path = self.project.codemeta_json_path
        with self._modified_pyproject_toml() as (tmp_ppt_path, name):
            tmp_dir = tmp_ppt_path.parent
            tmp_path = tmp_dir / codemeta_json_path.name
            with virtual_environment(update_pip=True, inherit=False) as (_venv_dir, venv_exe):
                cmd = [
                    venv_exe,
                    '-m', 'pip',
                    'install', '-U',
                    str(tmp_dir),
                    # TODO
                    # Remove setuptools once CodeMetaPy is updated for Python >
                    # 3.12
                    'CodeMetaPy', 'setuptools'
                ]
                LOGGER.debug('Running command: %s', cmd)
                subprocess.run(cmd, check=True)
                cmd = [
                    venv_exe,
                    '-m',
                    'codemeta.codemeta',
                    '--enrich',
                    '-O', str(tmp_path)
                ]
                LOGGER.debug('Running command: %s', cmd)
                subprocess.run(cmd, check=True, cwd=str(tmp_dir))
            with tmp_path.open('rb') as handle:
                codemeta = json.load(handle)
            codemeta['name'] = name
            codemeta['author'] = author_data
            codemeta['readme'] = self.project.git_repo.readme_url
            codemeta['maintainer'] = author_data[0]
            for cont in codemeta.get('contributor', []):
                if all(
                    cont[key] == author_data[0][key]
                    for key in ('familyName', 'givenName')
                ):
                    cont.clear()
                    cont.update(author_data[0])
            if version:
                codemeta['version'] = version
            codemeta_text = json.dumps(codemeta, indent=2, sort_keys=True)
            if not codemeta_text.endswith('\n'):
                codemeta_text += '\n'
            tmp_path.write_text(codemeta_text, encoding='utf-8')
            diff(tmp_path, codemeta_json_path)



#!/usr/bin/env python3
'''
File operations.
'''

import filecmp
import logging
import pathlib
import shutil
import subprocess
import tempfile


LOGGER = logging.getLogger(__name__)


def diff(path_1, path_2, differ='vimdiff'):
    '''
    Diff 2 paths.
    '''
    if not path_2.exists():
        shutil.copy(path_1, path_2)
        return
    if not filecmp.cmp(path_1, path_2, shallow=False):
        #  if path_1.read_bytes().rstrip(b'\n') == path_2.read_bytes().rstrip(b'\n'):
        #      return
        cmd = [differ, str(path_1), str(path_2)]
        LOGGER.info('Diffing %s and %s', path_1, path_2)
        subprocess.run(cmd, check=True)


def update_content(content, path, encoding='utf-8'):
    '''
    Interactively compare and merge new content.

    Args:
        content:
            The new content to merge.

        path:
            The target path.
    '''
    path = pathlib.Path(path).resolve()
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = pathlib.Path(tmp_dir)
        tmp_path = tmp_dir / path.with_stem('new').name
        tmp_path.write_text(content, encoding=encoding)
        diff(tmp_path, path)



#!/usr/bin/env python3
'''
Sphinx configuration file.
'''

# pylint: disable=invalid-name,redefined-builtin

import importlib
import pathlib
import sys

source_code = '../../src'
git_url = 'https://gitlab.inria.fr/jrye/None'

this_path = pathlib.Path(__file__).resolve()
sys.path.insert(0, str((this_path.parent / source_code).resolve()))

author = 'Jan-Michael Rye'
copyright = '2024, Inria'
project = 'Prometa'
html_theme = 'sphinx_rtd_theme'

autodoc_mock_imports = [
    'gitlab',
    'tomli_w',
]
extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo'
]
index_entries = []


def skip(_app, _what, name, _obj, would_skip, _options):  # pylint: disable=too-many-arguments
    '''Customize autodoc member skipping.'''
    if name == '__init__':
        return False
    return would_skip


def setup(app):
    '''Connect the skip function.'''
    app.connect('autodoc-skip-member', skip)


def linkcode_resolve(domain, info):
    '''Get source links for the linkcode extension.'''
    module = info['module']
    if domain != 'py' or not module:
        return None
    top_mod = importlib.import_module(module.split('.')[0])
    mod = importlib.import_module(module)
    top_mod_path = pathlib.Path(top_mod.__file__)
    mod_path = pathlib.Path(mod.__file__)
    subpath = str(mod_path.relative_to(top_mod_path.parent.parent))
    return f'{git_url}/-/blob/main/src/{subpath}'



#!/usr/bin/env python
'''
Project class.
'''

import logging

import spdx_matcher

from .citation import Citation
from .codemeta import CodeMeta
from .config import Config
from .gitlab.ci import GitlabCI
from .gitlab.repo import GitlabRepo
from .python.common import update_pyproject_toml, get_version
from .readme import ReadmeInserter


LOGGER = logging.getLogger(__name__)


class Project():
    '''
    Project class.
    '''
    def __init__(self, path, trust_commands=False):
        '''
        Args:
            path:
                The path to the project.
        '''
        self.git_repo = GitlabRepo(path)
        self.gitlab_ci = GitlabCI(self)
        self.config = Config(self.git_repo.path)
        self.trust_commands = trust_commands
        self.codemeta = CodeMeta(self)

    @property
    def readme_md_path(self):
        '''
        The README.md path.
        '''
        return self.git_repo.path / 'README.md'

    @property
    def pyproject_toml_path(self):
        '''
        The pyproject.toml path.
        '''
        return self.git_repo.path / 'pyproject.toml'

    @property
    def license_txt_path(self):
        '''
        The path to the LICENSE.txt file.
        '''
        return self.git_repo.path / 'LICENSE.txt'

    @property
    def spdx_license(self):
        '''
        The detected SPDX license name.
        '''
        path = self.license_txt_path
        try:
            detected, percent = spdx_matcher.analyse_license_text(path.read_text(encoding='utf-8'))
        except FileNotFoundError:
            LOGGER.error('No license file found at %s', path)
            return None
        licenses = list(detected['licenses'])
        if not licenses:
            LOGGER.error('Failed to detect license in %s', path)
            return None
        lic = licenses[0]
        if percent < 0.9:
            LOGGER.warning(
                'Detected %s license in %s but certainty is only %(0.0f) %',
                lic,
                path,
                percent * 100.
            )
        return lic

    @property
    def codemeta_json_path(self):
        '''
        The path to the codemeta.json file.
        '''
        return self.git_repo.path / 'codemeta.json'

    @property
    def citation_cff_path(self):
        '''
        The path to the CITATION.cff file.
        '''
        return self.git_repo.path / 'CITATION.cff'

    @property
    def urls(self):
        '''
        A dict of URLs for the project.
        '''
        host, namespace, name = self.git_repo.parsed_origin
        homepage = f'https://{host}/{namespace}/{name}'

        urls = {}
        urls['Homepage'] = homepage
        urls['Source'] = f'{homepage}.git'
        if self.gitlab_ci.data.get('pages'):
            pages_fmt = self.config.get('gitlab', 'pages_urls', host)
            if pages_fmt:
                urls['Documentation'] = pages_fmt.format(namespace=namespace, name=name)
        urls['Issues'] = f'{homepage}/-/issues'
        return urls

    def update(self):
        '''
        Update project metadata.
        '''
        if self.pyproject_toml_path.exists():
            # TODO
            # Find a better way to automate version management while updating
            # the files. The issue is that prometa has to modify several files
            # but these modifications are visible to the VCS. This causes tools
            # such as setuptools-scm to miscalculate the version because it
            # doesn't know that the changes will be merged into the previous
            # commit.
            version = get_version(self.git_repo.path)
            update_pyproject_toml(self)
            self.codemeta.from_pyproject_toml(version=version)

        if self.codemeta_json_path.exists():
            Citation(self).update()

        readme_inserter = ReadmeInserter(self)
        readme_inserter.update(self.readme_md_path)

        self.gitlab_ci.update()

    @property
    def name(self):
        '''
        The project name. It will use the value in codemeta.name if it exists,
        otherwise it will use the name of the project's directory.
        '''
        name = self.codemeta.name
        if name is not None:
            return name
        return self.git_repo.path.name



#!/usr/bin/env python
'''
Update the CI file.
'''

import logging
import re

import yaml

from ..file import update_content

LOGGER = logging.getLogger(__name__)


class GitlabCI():
    '''
    Wrapper around the gitlab-ci file.
    '''
    def __init__(self, project, path='.gitlab-ci.yml'):
        '''
        Args:
            project:
                A Project instance.

            path:
                The path to the gitlab-ci file, relative to the repository root
                directory.
        '''
        self.project = project
        self.path = project.git_repo.path / path
        self._data = None

    @property
    def data(self):
        '''
        The CI configuration data.
        '''
        if self._data is None:
            self._data = self.load()
        return self._data

    def load(self):
        '''
        Load the file data.

        Returns:
            The loaded data, or an empty dict if the file does not exist.
        '''
        try:
            with self.path.open('rb') as handle:
                return yaml.safe_load(handle)
        except FileNotFoundError:
            return {}

    def update_pages(self):
        '''
        Update the pages job.
        '''
        pages = self.data.get('pages')
        if not pages:
            return
        pages['artifacts'] = {'paths': ['public']}
        pages['only'] = ['main']
        pages['stage'] = 'deploy'

    def add_register_pip_pkg(self):
        '''
        Add a jot to register a pip package.
        '''
        key = 'register_pip_pkg'
        if self.project.pyproject_toml_path.exists():
            self.data[key] = {
                'image': 'python:latest',
                'only': ['main'],
                'script': [
                    'pip install build twine',
                    'python -m build',
                    'TWINE_PASSWORD=${CI_JOB_TOKEN} TWINE_USERNAME=gitlab-ci-token '
                    'python -m twine upload --repository-url '
                    '${CI_API_V4_URL}/projects/${CI_PROJECT_ID}/packages/pypi '
                    'dist/*'
                ],
                'stage': 'deploy'
            }
        else:
            self.data.pop(key)

    def add_release_job(self):
        '''
        Add a release job that triggers when Git tags are pushed.
        '''
        key = 'release_job'
        regex = self.project.config.get('gitlab', 'release_tag_regex')
        if regex:
            self.data[key] = {
                'image': 'registry.gitlab.com/gitlab-org/release-cli:latest',
                'release': {
                    'description': 'Release $CI_COMMIT_TAG',
                    'tag_name': '$CI_COMMIT_TAG'
                },
                'rules': [{'if': f'$CI_COMMIT_TAG =~ /{regex}/'}],
                'script': ['echo "running release_job"'],
                'stage': 'release'
            }
        else:
            self.data.pop(key)

    def _manage_swh_hook(self, gproj):
        '''
        Manage the SWH webhooks.

        Args:
            gproj:
                The python-gitlab Project instance.
        '''
        # Reset to releases_events when SoftwareHeritage updates their API.
        #  events_key = 'releases_events'
        events_key = 'tag_push_events'
        url = 'https://archive.softwareheritage.org/api/1/origin/save/webhook/gitlab/'

        swh_hook = None
        add = self.project.codemeta_json_path.exists()

        for hook in gproj.hooks.list(iterator=True):
            if hook.attributes['url'] == url:
                if add and swh_hook is None:
                    swh_hook = hook
                else:
                    LOGGER.info('Deleting SWH hook for %s.', gproj.name)
                    hook.delete()

        if add and swh_hook is None:
            LOGGER.info('Creating SWH webhook for %s', gproj.name)
            swh_hook = gproj.hooks.create({
                'url': url,
                events_key: True
            })

        if swh_hook:
            changed = False
            for key, value in swh_hook.attributes.items():
                if key.endswith('_events'):
                    expected_value = key == events_key
                    if value != expected_value:
                        changed = True
                        setattr(swh_hook, key, expected_value)
            if changed:
                LOGGER.info('Updating SWH hook for %s.', gproj.name)
                swh_hook.save()

    def manage_hooks(self):
        '''
        Add or remove hooks depending on the current repository configuration.
        '''
        with self.project.config.gitlab as glab:
            glab.auth()
            gproj = self.project.git_repo.get_gitlab_project(glab)
            self._manage_swh_hook(gproj)

    def add_stages(self):
        '''
        Add the list of stages for all added jobs.
        '''
        recognized_stages = [
            '.pre',
            'build',
            'test',
            'release',
            'deploy',
            '.post'
        ]
        stages = set()
        for key, value in self.data.items():
            if not isinstance(value, dict):
                continue
            stage = value.get('stage')
            if stage is None:
                LOGGER.error('No stage specified for job %s in %s', key, self.path)
                continue
            stages.add(stage)
        self.data['stages'] = [stage for stage in recognized_stages if stage in stages]

    def add_tags(self):
        '''
        Add runner tags. This will deduplicate tags and also ensure that jobs
        using the same tags reference each other in the YAML output.
        '''
        tag_map = self.project.config.get('gitlab', 'ci_tags')
        if tag_map:
            tag_map = tuple((re.compile(regex), tags) for regex, tags in tag_map.items())
            previous_sets = []
            for name, data in self.data.items():
                if not isinstance(data, dict):
                    continue
                collected_tags = set()
                for regex, new_tags in tag_map:
                    if regex.search(name):
                        collected_tags.update(new_tags)
                collected_tags = sorted(collected_tags)
                # PyYAML seems to only emit references when nested items refer
                # to the same instance so ensure that identical lists do.
                for prev in previous_sets:
                    if collected_tags == prev:
                        collected_tags = prev
                        break
                else:
                    previous_sets.append(collected_tags)
                self.data[name]['tags'] = collected_tags

    def update(self):
        '''
        Update the CI file by detecting which jobs should be added.
        '''
        self.update_pages()
        self.add_register_pip_pkg()
        self.add_release_job()
        self.add_stages()
        self.add_tags()
        update_content(yaml.dump(self.data), self.path)
        self.manage_hooks()
